#include <cassert>

#include "graph.h"
#include "snapshot.h"

// TODO: after optimizing the memory layout maybe enforce that rolling back actually maintains the actual positions in the vectors, should be relatively easy

/* change the degree of a vertex
 * takes care of placing v in the correct lists
 */
void vert_change_deg(Graph &G, Vertex *v, size_t new_deg) {
	if (v->deg == new_deg)
		return;

	if (v->deg == 0) {
		G.V.push_back(v);
		v->iterV = G.V.size() - 1;
	}


	if (new_deg == 0) {
		size_t idx = v->iterV;
		assert(G.V[idx] == v);

		swap(G.V[idx], G.V.back());
		G.V[idx]->iterV = idx;
		G.V.resize(G.V.size() - 1);
		v->iterV = -1;
	}

	v->deg = new_deg;
}

void register_edge_deletion(Graph &G, Edge *e) {
	GM_Edge_Deletion *last = nullptr;
	bool merged = false;

	if (G.changes.size() != 0) {
		// if the last transformation was an edge deletion merge them
		last = dynamic_cast<GM_Edge_Deletion *> (G.changes.back());
		merged = last != nullptr;
	}

	if (last == nullptr) {
		last = new GM_Edge_Deletion();
	}

	last->edges.push_back(e);
	if (!merged) {
		G.changes.push_back(last);
	}
}

void delete_edge(Graph &G, Edge *e) {
	Vertex *a = e->end[0];
	Vertex *b = e->end[1];

	assert(a->edges[e->pos[0]].first == b);
	assert(b->edges[e->pos[1]].first == a);

	// delete the edge from the edgelist
	a->edges[e->pos[0]] = a->edges.back();
	size_t idx = e->pos[0];
	Edge *e_tmp = a->edges.back().second;
	int p = e_tmp->end[0] == a ? 0 : 1;
	e_tmp->pos[p] = idx;
	a->edges.resize(a->edges.size() - 1);

	b->edges[e->pos[1]] = b->edges.back();
	idx = e->pos[1];
	e_tmp = b->edges.back().second;
	p = e_tmp->end[0] == b ? 0 : 1;
	e_tmp->pos[p] = idx;
	b->edges.resize(b->edges.size() - 1);

	// move edge to the vertex's covered edges
	a->covered.push_back(make_pair(b, e));
	b->covered.push_back(make_pair(a, e));
	e->pos[0] = a->covered.size() - 1;
	e->pos[1] = b->covered.size() - 1;

	// add the edge to the covered edges
	register_edge_deletion(G, e);
	G.VDS.E.push_back(e);
	assert(G.E[e->iterE] == e);
	G.E[e->iterE] = G.E.back();
	G.E[e->iterE]->iterE = e->iterE;
	e->iterE = -1;
	G.E.resize(G.E.size() - 1);

	e->covered = true;

	// update degrees
	vert_change_deg(G, a, a->edges.size());
	vert_change_deg(G, b, b->edges.size());
}

/* deletes a vertex from the graph
 * deleted edges will be treated like covered edges
 */
void delete_vertex(Graph &G, Vertex *a) {
	// cover the edges
	while (a->edges.size() > 0) {
		// TODO: back vs front ?
		Edge *e = a->edges.back().second;

		delete_edge(G, e);
	}
}

/* add a vertex to the vertex cover
 * 
 * this function also does all the housekeeping for making sure operations can be reversed
 */
void add_to_vds(Graph &G, Vertex *v) {
	G.VDS.cost += v->weight;

	delete_vertex(G, v);
	G.VDS.S.push_back(v);
}

struct graph_snapshot create_snapshot(Graph &G) {
	struct graph_snapshot snapshot;

	snapshot.vds.num_vertices = G.VDS.S.size();
	snapshot.vds.num_edges    = G.VDS.E.size();
	snapshot.vds.cost         = G.VDS.cost;
	snapshot.num_changes      = G.changes.size();

	// create first graph modification entry, to prevent erroneous auto-merging
	G.changes.push_back(new GM_Edge_Deletion());

	return snapshot;
}




/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/




/* mark an edge as uncovered - move the edge to the appropriate lists */
void uncover_edge(Graph &G, Edge *e) {
	assert(e->covered == true);
	e->covered = false;

	// add the edge back to the vertices
	for (int i = 0; i < 2; i++) {
		Vertex *v = e->end[i];
		size_t idx = e->pos[i];
		v->edges.push_back(v->covered[idx]);

		assert(idx == v->covered.size() - 1);
		v->covered.resize(v->covered.size() - 1);

		e->pos[i] = v->edges.size() - 1;
	}
	Vertex *a = e->end[0];
	Vertex *b = e->end[1];

	assert(a->edges[e->pos[0]].first == b);
	assert(b->edges[e->pos[1]].first == a);
	assert(a->deg + 1 == a->edges.size());
	assert(b->deg + 1 == b->edges.size());

	// update degrees
	vert_change_deg(G, a, a->deg + 1);
	vert_change_deg(G, b, b->deg + 1);

	// add the edge back into the graph
	G.E.push_back(e);
	e->iterE = G.E.size() - 1;
}

void GM_Edge_Deletion::undo(Graph &G) {
	
	for (size_t i = 0; i < edges.size(); i++) {
		size_t j = edges.size() - i - 1;
		uncover_edge(G, edges[j]);
	}
}

void GM_Edge_Deletion::translate(Graph &G, list<Vertex *> &sol) {
	// nothing to do
	return;
}

void restore_snapshot(Graph &G, struct graph_snapshot &s)
{
	// unmark picked vertices
	for (size_t i = s.vds.num_vertices; i < G.VDS.S.size(); i++) {
		assert(G.VDS.S[i]->cstr_in_sol);
		G.VDS.S[i]->cstr_in_sol = false;
	}
	G.VDS.S.erase(G.VDS.S.begin() + s.vds.num_vertices, G.VDS.S.end());

	// undo graph transformations
	for (size_t i = 0; i < G.changes.size() - s.num_changes; i++) {
		size_t j = G.changes.size() - i - 1;
		GraphModification *op = G.changes[j];
		op->undo(G);
		delete op;
	}
	G.changes.erase(G.changes.begin() + s.num_changes, G.changes.end());

	// erase the edges from the VC
	G.VDS.E.erase(G.VDS.E.begin() + s.vds.num_edges, G.VDS.E.end());

	G.VDS.cost = s.vds.cost;
}
