#include <utility>
#include <cassert>
#include <algorithm>

#include "graph.h"
#include "branch.h"
#include "read_sol.h"
#include "optimize.h"
#include "stats.h"
#include "constraints.h"
#include "disjoint_bound.h"
#include "cut_bound.h"

bool vds_branch(Graph &G, long long k);

bool vds_branch_split(Graph &G, size_t components, long long k) {
	if (k < 0)
		return false;
	G.cuts++;

	struct graph_snapshot snapshot = create_snapshot(G);


	// every node is part of a component
	for(Vertex *v: G.V) {
		assert(v->component != 0);
	}

	size_t old_v_size = G.V.size();
	size_t old_e_size = G.E.size();
	vector<GraphModification *> whole_changes;
	vector<Vertex *> whole_VDS;
	long VDS_cost = G.VDS.cost;

	whole_changes.swap(G.changes);
	whole_VDS.swap(G.VDS.S);

	struct Component {
		vector<Vertex *> V;
		vector<Edge *> E;

		bool operator<(const Component &rhs) {
			return V.size() < rhs.V.size();
		}
	};

	struct Component C[components+1];

	for (Vertex *v: G.V) {
		size_t i = v->component;
		C[i].V.push_back(v);
	}

	for (Edge *e: G.E) {
		size_t i = e->end[0]->component;
		C[i].E.push_back(e);
	}

	G.V.clear();
	G.E.clear();

	bool success = true;

	// TODO: solve smallest component first
	sort(&C[1], &C[components+1]);

	size_t first_fail = 100000;
	for (size_t i = 1; i <= components; i++) {
		for (Vertex *v: C[i].V) {
			G.V.push_back(v);
			v->iterV = G.V.size() - 1;
		}
		for (Edge *e: C[i].E) {
			G.E.push_back(e);
			e->iterE = G.E.size() - 1;
		}
		G.VDS.cost = 0;

		assert(G.V.size() != 0);
		assert(G.VDS.S.size() == 0);
		assert(G.changes.size() == 0);
		// dubug snapshot
		struct graph_snapshot ds = create_snapshot(G);
		vector<Vertex *> debug_V{G.V.begin(), G.V.end()};


		long long lower_bound = disjoint_bound(G);
		long long c_k = lower_bound;

		if (i == components)
			c_k = k;

		bool found = false;
		while (c_k <= k) {
			found = vds_branch(G, c_k);

			if (found)
				break;
			c_k = c_k + 1;
		}

		if (!found) {
			assert(G.VDS.S.size() == 0);
			assert(G.changes.size() == 1);
			restore_snapshot(G, ds);
			assert(G.changes.size() == 0);
			assert(G.V.size() == C[i].V.size());
			assert(G.E.size() == C[i].E.size());
		}
		else {
			/* G is now a 2-club cluster graph */
			while (!G.V.empty()) {
				Vertex *v = G.V.front();
				delete_vertex(G, v);
			}
		}

		k -= c_k;
		VDS_cost += c_k;

		whole_changes.insert(whole_changes.end(), G.changes.begin(), G.changes.end());
		G.changes.clear();
		whole_VDS.insert(whole_VDS.end(), G.VDS.S.begin(), G.VDS.S.end());
		G.VDS.S.clear();


		if (!found) {
			for (size_t j = i+1; j <= components; j++) {
				G.V.insert(G.V.end(), C[j].V.begin(), C[j].V.end());
				G.E.insert(G.E.end(), C[j].E.begin(), C[j].E.end());
			}

			for (size_t i = 0; i < G.V.size(); i++) {
				G.V[i]->iterV = i;
			}
			for (size_t i = 0; i < G.E.size(); i++) {
				G.E[i]->iterE = i;
			}

			success = false;
			break;
		}

		G.V.clear();
		G.E.clear();
	}


	assert(G.changes.size() == 0);
	assert(G.VDS.S.size() == 0);

	if (success) {
		assert(G.V.size() == 0);
		assert(G.E.size() == 0);
	}

	// cleanup

	G.changes.swap(whole_changes);
	G.VDS.S.swap(whole_VDS);
	G.VDS.cost = VDS_cost;

	if (!success) {
		// undo solutions to the components  that were solved
		restore_snapshot(G, snapshot);

		for (Vertex *v: G.V)
			v->marked = true;
		bool quit = false;
		for (size_t i = 1; i <= components; i++) {
			size_t missing = 0;
			for (Vertex *v: C[i].V) {
				assert(v->marked);
				if (!v->marked)
					missing++;
			}
			if (missing > 0) {
				cout << i << " is missing " << missing << endl;
				quit = true;
			}
		}

		assert(!quit);

		for (Vertex *v: G.V)
			v->marked = false;
		assert(old_v_size == G.V.size());
		assert(old_e_size == G.E.size());
	}

	return success;
}

void component_dfs(Vertex* v, size_t component_id) {
	v->component = component_id;
	for (auto edge : v->edges) {
		Vertex *u = edge.first;
		if (u->component == 0) {
			component_dfs(u, component_id);
		}
	}
}

void find_p4(Graph &G, Vertex *&bs, Vertex *&bt, Vertex *&bu, Vertex *&bv) {

	long long best_score = 0;
	bs = bt = bu = bv = nullptr;

	for (Vertex *a: G.V) {
		vector<Vertex *> marked;
		vector<Vertex *> current;
		vector<Vertex *> next;

		assert(!a->marked);
		a->marked = true;
		a->dist   = 0;
		a->tmp    = nullptr;
		marked.push_back(a);
		current.push_back(a);


		// BFS
		for (int dist = 0; dist < 3; dist++) {
			for (Vertex *b: current) {
				assert(b->dist == dist);
				assert(b->marked);

				b->score = b->permanent ? 1000000ll : b->weight;
				if (dist != 0)
					b->score += b->tmp->score;

				for (auto edge: b->edges) {
					Vertex *c = edge.first;

					if (c->marked) {
						if (c == a)
							continue;
						assert(b->dist >= c->tmp->dist);
						// allow to set the parent of c to b if b is better than c->tmp
						if (b->dist != c->tmp->dist || b->score < c->tmp->score)
							continue;
					}

					c->tmp = b;
					c->dist = dist+1;

					if (!c->marked) {
						c->marked = true;
						next.push_back(c);
						marked.push_back(c);
					}
				
				}
			}

			current.clear();
			swap(current, next);
		}

		for (Vertex *v: current) {
			assert(v->dist == 3);
			Vertex *u = v->tmp;
			Vertex *t = u->tmp;
			Vertex *s = t->tmp;
			assert(s == a);

			long long score = 0;
			score += s->permanent ? 1000000ll : s->weight;
			score += t->permanent ? 1000000ll : t->weight;
			score += u->permanent ? 1000000ll : u->weight;
			score += v->permanent ? 1000000ll : v->weight;

			if (score > best_score) {
				bs = s;
				bt = t;
				bu = u;
				bv = v;
				best_score = score;
			}
		}

		// cleanup
		for (Vertex *b: marked) {
			assert(b->marked);
			b->marked = false;
			b->tmp = nullptr;
			b->dist = 0;
			b->score = 0;
		}

		for (Vertex *b: G.V)
			assert(!b->marked);
		marked.clear();
		current.clear();
		next.clear();
	}

	return;
} 

bool pair_cmp(pair<Vertex *, Edge *> &e1, pair<Vertex *, Edge *> &e2) {
	return e1.first < e2.first;
}

// returns true if a VDS has been found, if not false and a lower bound is returned
bool vds_branch(Graph &G, long long k) {
	Vertex *s = nullptr;
	Vertex *t = nullptr;
	Vertex *u = nullptr;
	Vertex *v = nullptr;



#if 0
#if 0
	//random_shuffle(G.V.begin(), G.V.end());

	// sort the pointers for better memory access
	sort(G.V.begin(), G.V.end());
#endif
	for (size_t i = 0; i < G.V.size(); i++) {
		Vertex *v = G.V[i];
		v->iterV = i;

#if 0
		sort(v->edges.begin(), v->edges.end(), pair_cmp);
		for (size_t j = 0; j < v->edges.size(); j++) {
			Edge *e = v->edges[j].second;
			if (e->end[0] != v)
				e->flip();
			e->pos[0] = j;
		}
#endif
	}
#endif

	find_p4(G, s, t, u, v);
	Vertex *P[4] = {s, t, u, v};

	long long lower_bound = disjoint_bound(G);
	if (k < 0) {
		return false;
	}

	G.recursive_steps++;

	// check constraints
	for (size_t i = 0; i < G.constraints.size(); i++) {
		if (!G.constraints[i]->check(G)) {
			return false;
		}
	}

	if (s == nullptr) {
		/* it is possible to overshoot k if we are solving multiple components
		 * If an optimal solution for one of the components breaks one of our
		 * constraints then we'll just look for a different one, with likely a
		 * higher value of k. Because we can't guarantee that the same reduction
		 * rules will be executed that enforce constraints that might mean that
		 * we will find the same or similar solution that is again optimal for
		 * the component but breaks constraints, and we will have some budget left
		 */ 
		//assert(k == 0);		// we must never overshoot k
		return true;
	}

	long long old_k = k;
	long long min_bound;

	assert(lower_bound >= 1);

	if (false) {
		long long min_deg = G.V.front()->deg;
		for (Vertex *a: G.V) {
			min_deg = min(min_deg, (long long) a->deg);
		}

		if (min_deg > lower_bound) {
			cout << "# lower bound could perhaps be improved from " << lower_bound << " to " << min_deg << endl;
		}
	}

	if (false) {
		long long lb_tmp = 10000000;
		for (Vertex *a: G.V)
			lb_tmp = min(lb_tmp, a->weight);

		lower_bound = lb_tmp;
	}

	if (k < lower_bound) {
		goto fail;
	}

	/* find components in the graph and handle them seperately */
	if (true) {
		size_t components = 0;
		for (Vertex *a: G.V) {
			a->component = 0;
		}

		for (Vertex *a: G.V) {
			if (a->component != 0)
				continue;
			components++;
			component_dfs(a, components);
		}

		if (components > 1)
			return vds_branch_split(G, components, k);
	}

	lower_bound = max(lower_bound, cut_bound(G));
	if (k < lower_bound) {
		goto fail;
	}

	/* do the branching */

	for (int i = 0; i < 4; i++) {
		if (P[i]->permanent)
			continue;

		/* information for rollback */
		struct graph_snapshot snapshot;
		snapshot = create_snapshot(G);
		size_t old_vds_s       = G.VDS.S.size();
		size_t old_vds_e       = G.VDS.E.size();
		size_t old_e           = G.E.size();
		size_t old_v           = G.V.size();
		size_t old_constraints = G.constraints.size();
		size_t old_changes     = G.changes.size() - 1;

		for (int j = 0; j < i; j++)
			mark_permanent(G, P[j]);

		add_to_vds(G, k, P[i]);


		/* optimize */
		graph_optimize(G, k);	// can decrement k

		/* branch */

		bool found = vds_branch(G, k);
		if (found)
			return true;

		/* rollback */
		k = old_k;

		restore_snapshot(G, snapshot);
		assert(G.VDS.S.size()       == old_vds_s);
		assert(G.VDS.E.size()       == old_vds_e);
		assert(G.E.size()           == old_e);
		assert(G.V.size()           == old_v);
		assert(G.constraints.size() == old_constraints);
		assert(G.changes.size()     == old_changes);
	}	

fail:

	return false;
}

void two_club_vertex_delete(Graph &G) {
	long long k = max(0ll, (long long) G.V.size() - 3); // upper bound
	list<Vertex *> solution;

	// always check that no permanent vertex was picked
	G.constraints.push_back(new NoPermanentVertexConstraint());

	/* optimize graph */
	for (int i = 0; i < 5; i++)
		graph_optimize(G, k);

	/* reallocate the graph but packed more efficiently */
	Graph G2;
	G2.V.reserve(G.V.size());
	G2.E.reserve(G.E.size());

	vector<Vertex> vertex_pool;
	vector<Edge> edge_pool;

	vertex_pool.reserve(G.V.size());
	edge_pool.reserve(G.E.size());

	for (Vertex *v: G.V) {
		vertex_pool.emplace_back(Vertex(v->id, v->name));
		v->tmp = &vertex_pool[vertex_pool.size() - 1];
		vertex_pool.back().permanent = v->permanent;
		vertex_pool.back().weight = v->weight;
	}
	for (Edge *e: G.E) {
		edge_pool.emplace_back(G2, e->id, e->end[0]->tmp, e->end[1]->tmp);
		Edge *e2 = &edge_pool[edge_pool.size() - 1];
		assert(e2->end[0]->edges[e2->pos[0]].second == e2);
	}

	G2.constraints.push_back(new NoPermanentVertexConstraint());
	print_graph_optimization(G);

	// save the current VC into a different vector
	// this has to be done because vc_branch assumes that when it is first
	// called G.VC.V is empty
	//swap(G.VDS.S, G.VDS.S_backup);

	k = disjoint_bound(G2);
	k = max(k, cut_bound(G2));

	cout << "# initial lower bound k = " << k << endl;


	while (true) {
		assert(G2.VDS.S.size() == 0);
		bool found = vds_branch(G2, k);

		if (found)
			break;
		k = k + 1;

		cout << "# k = " << k << endl;
	}

	

	// add the old VDS back to the current one
	//G.VDS.S_backup.insert(G.VDS.S_backup.end(), G.VDS.S.begin(), G.VDS.S.end());
	//swap(G.VDS.S, G.VDS.S_backup);

	// read off the VDS for the original graph from the transformed graph
	read_sol(G2, solution);
	list<Vertex *> solution2;

	for (Vertex *v: solution) {
		assert(G.name_map.find(v->name) != G.name_map.end());
		rs_add_vertex(G.name_map[v->name], solution2);
	}
	read_sol(G, solution2);

	cout << "# VDS size = " << solution2.size() << endl;
	cout << "# recursive steps: " << G2.recursive_steps << endl;
	cout << "# cuts: " << G2.cuts << endl;

	for (Vertex *v: solution2) {
		cout << v->name << endl;
	}
}
