#include <cassert>
#include <algorithm>

#include "graph.h"
#include "disjoint_bound.h"

static bool vertex_compare(Vertex *a, Vertex *b) {
	return a->deg < b->deg;
}


bool disjoint_bound_find_p4(Graph &G, Vertex *a, long long &bound) {
	vector<Vertex *> marked;
	vector<Vertex *> current;
	vector<Vertex *> next;
	Vertex *s = nullptr;
	Vertex *t = nullptr;
	Vertex *u = nullptr;
	Vertex *v = nullptr;

	bool found = false;
	if (a->overlap == 0)
		return false;

	marked.push_back(a);
	current.push_back(a);
	a->marked = true;
	a->tmp = nullptr;
	a->dist = 0;

	// BFS
	for (int dist = 0; dist < 3; dist++) {
		for (Vertex *b: current) {
			if (dist == 0)
				b->score = b->deg;
			else
				b->score = b->tmp->score + b->deg;

			assert(dist == 0 || b->tmp != nullptr);

			if (dist == 2) {
				if (b->overlap == 0 || b->tmp->overlap == 0)
					continue;
			}

			for (auto edge: b->edges) {
				Vertex *c = edge.first;
				if (c == a || (dist == 2 && c->overlap == 0))
					continue;

				if (c->marked) {
					assert(b->dist >= c->tmp->dist);
					if (b->dist != c->tmp->dist || b->score > c->tmp->score)
						continue;
				}

				if (!c->marked) {
					marked.push_back(c);
					next.push_back(c);
				}
				c->marked = true;
				c->tmp     = b;
				c->dist    = dist+1;
			}
		}

		current.clear();
		swap(current, next);
	}

	Vertex *best = nullptr;
	long long best_score = 1000000;
	for (Vertex *b: current) {
		long long score = 0;
		v = b;
		u = v->tmp;
		t = u->tmp;
		s = t->tmp;
		score = s->deg + t->deg + u->deg + v->deg;
		if (score < best_score || best == nullptr) {
			best = b;
			best_score = score;
		}
		
	}

	if (best != nullptr) {
		v = best;
		u = v->tmp;
		t = u->tmp;
		s = t->tmp;
		assert(s == a);
		assert(v->dist == 3);
		assert(u->dist == 2);
		assert(t->dist == 1);
		assert(s->dist == 0);
		long long overlap = min(min(s->overlap, t->overlap), min(u->overlap, v->overlap));
		s->overlap -= overlap;
		t->overlap -= overlap;
		u->overlap -= overlap;
		v->overlap -= overlap;

		found = true;
		bound += overlap;
	}

	// cleanup
	for (Vertex *b: marked) {
		assert(b->marked);
		b->marked = false;
		b->tmp = nullptr;
		b->dist = 0;
		b->score = 0;
	}

	for (Vertex *b: G.V)
		assert(!b->marked);
	marked.clear();
	current.clear();
	next.clear();

	return found;
} 

size_t DISJOINT_BOUND_SHUFFLE_PCT  = 50;
float  DISJOINT_BOUND_SHUFFLE_DIST = 0.5;

void disjoint_bound_shuffle(vector<Vertex *> &array) {

	for (long long i = 0; i < array.size(); i++) {
		if (rand() % 100 >= DISJOINT_BOUND_SHUFFLE_PCT)
			continue;

		long long dist = (long long) ((rand() % array.size()) * DISJOINT_BOUND_SHUFFLE_DIST);
		if (rand() % 2 == 0)
			dist = -dist;

		if (0 <= i + dist && i + dist < array.size())
			swap(array[i], array[i + dist]);
	}
}

#define NUM_ITER 1
long long disjoint_bound(Graph &G) {
	long long bound[NUM_ITER];

	// do some sort of sorting
	vector<Vertex *> V(G.V);
	sort(V.begin(), V.end(), vertex_compare);

	for (Vertex *a: G.V) {
		a->overlap = a->permanent ?  1000000ll : a->weight;
	}

	bound[0] = 0;
	for (Vertex *a: V) {
		if (a->marked)
			continue;

		while (disjoint_bound_find_p4(G, a, bound[0]))
			;
	}

	vector<Vertex *> V_tmp{G.V.begin(), G.V.end()};
	for (int i = 1; i < NUM_ITER; i++) {
		disjoint_bound_shuffle(V_tmp);
		bound[i] = 0;

		for (Vertex *a: G.V) {
			a->overlap = a->permanent ?  1000000ll : a->weight;
		}
		for (Vertex *a: V_tmp) {
			if (a->marked)
				continue;

			while (disjoint_bound_find_p4(G, a, bound[i]))
				;
		}
	}

	long long max_bound = bound[0];
	for (int i = 1; i < NUM_ITER; i++)
		max_bound = max(max_bound, bound[i]);

	return max_bound;
}
