#include <vector>
#include <cassert>
#include <string>
#include <iostream>
#include <stack>

#include "flow.h"

void F_Edge::send_flow(long flow) {
	cap -= flow;
	rev->cap += flow;

	assert(cap >= 0 && rev->cap >= 0);
}

F_Vertex::F_Vertex() {
}

F_Edge::F_Edge() {
}

static stack<F_Vertex *, vector<F_Vertex *> > vertex_stack;
static stack<F_Edge *, vector<F_Edge *> > edge_stack;

static F_Vertex *get_vertex_from_stack() {
	F_Vertex *v;
	if (!vertex_stack.empty()) {
		v = vertex_stack.top();
		vertex_stack.pop();
		assert(v->edges.size() == 0);
		return v;
	}
	else	return new F_Vertex();
}

static F_Edge *get_edge_from_stack() {
	F_Edge *e;
	if (!edge_stack.empty()) {
		e = edge_stack.top();
		edge_stack.pop();
		return e;
	}
	else	return new F_Edge();
}

F_Vertex *create_vertex(F_Graph &G) {
	F_Vertex *v = get_vertex_from_stack();

	v->marked = false;
	v->prev = nullptr;

	G.V.push_back(v);

	return v;
}

F_Edge *create_edge(F_Graph &G, F_Vertex *a, F_Vertex *b, long cap, Edge *dual) {
	F_Edge *e1 = get_edge_from_stack();
	F_Edge *e2 = get_edge_from_stack();

	e1->a = a;
	e1->b = b;
	e2->a = b;
	e2->b = a;

	e1->dual = dual;
	e2->dual = nullptr;
	e1->rev = e2;
	e2->rev = e1;
	e1->cap = cap;
	e2->cap = 0;

	a->edges.push_back(make_pair(b, e1));
	b->edges.push_back(make_pair(a, e2));

	G.E.push_back(e1);
	G.E.push_back(e2);

	return e1;
}

void F_Graph::clear() {
	for (F_Vertex *v: V) {
		v->edges.clear();
		vertex_stack.push(v);
	}
	for (F_Edge *e: E)
		edge_stack.push(e);

	V.clear();
	E.clear();
}

F_Graph::~F_Graph() {
	clear();
}


/******************************************************************************/
/******************************************************************************/
/******************************************************************************/

long dfs(F_Vertex *v, F_Vertex *t, long cap) {
	if (v == t)
		return cap;

	v->marked = true;

	for (auto edge: v->edges) {
		F_Vertex *a = edge.first;
		if (a->marked || edge.second->cap == 0)
			continue;

		long cap2 = edge.second->cap;
		long flow = dfs(a, t, min(cap, cap2));
		if (flow > 0) {
			edge.second->send_flow(flow);
			return flow;
		}
	}

	return 0;
}

void bfs(F_Graph &G, F_Vertex *s, F_Vertex *t) {
	static queue<F_Vertex *> Q;

	for (F_Vertex *v: G.V) {
		v->prev = nullptr;
		v->marked = false;
	}

	Q.push(s);
	s->marked = true;
	
	while (!Q.empty()) {
		F_Vertex *v = Q.front();
		Q.pop();

		for (auto edge: v->edges) {
			F_Vertex *a = edge.first;
			F_Edge *e = edge.second;
			if (a->marked || e->cap == 0)
				continue;

			a->marked = true;
			a->prev = e;
			if (a == t)
				goto done;
			Q.push(a);
		}
	}
done:
	while(!Q.empty())
		Q.pop();
	return;
}

long compute_max_flow(F_Graph &G, F_Vertex *s, F_Vertex *t) {
	bool found;
	long max_flow = 0;
	do {
#if 0
		found = false;
		s->marked = true;
		for (auto edge: s->edges) {
			F_Vertex *a = edge.first;
			long cap = edge.second->cap;
			if (cap == 0)
				continue;
			
			long flow = dfs(a, t, cap);
			if (flow > 0) {
				found = true;
				max_flow += flow;
				edge.second->send_flow(flow);
			}
		}
	} while (found);
#else
		bfs(G, s, t);
		if (t->prev == nullptr)
			break;

		F_Edge *e = t->prev;
		long flow = e->cap;
		while (e != nullptr) {
			flow = min(flow, e->cap);
			e = e->a->prev;
		}

		e = t->prev;
		while (e != nullptr) {
			e->send_flow(flow);
			e = e->a->prev;
		}
		max_flow += flow;

	} while (true);
#endif

	long incoming = 0;
	for (auto edge: s->edges) {
		incoming += edge.second->rev->cap;
	}
	assert(incoming == max_flow);
	incoming = 0;
	for (auto edge: t->edges) {
		incoming += edge.second->cap;
	}
	assert(incoming == max_flow);

	return max_flow;
}
