#include <vector>
#include <cassert>

#include "graph.h"
#include "flow.h"
#include "kplus_disjoint.h"


void find_neighbourhood_helper(vector<Vertex *> &A, vector<Vertex *> &N) {
	for (Vertex *a: A) {
		for (auto edge: a->edges) {
			Vertex *b = edge.first;
			if (b->marked)
				continue;
			b->marked = true;
			N.push_back(b);
		}
	}
}


bool tc_kplus_disjoint_single(Graph &G, long long &k, Vertex *v) {
	bool success = false;
	F_Graph Gf;
	F_Vertex *s = create_vertex(Gf);
	F_Vertex *t = create_vertex(Gf);
	long flow;

	if (k <= 0)
		return false;

	vector<Vertex *> tmp;
	tmp.push_back(v);

	vector<Vertex *> N1; // dist 1 to v
	vector<Vertex *> N2; // dist 2 to v
	vector<Vertex *> N3; // dist 3 to v

	assert(!v->marked);
	v->marked = true;
	v->dist = 0;
	find_neighbourhood_helper(tmp,N1);
	long N1_weight = 0;
	long N2_weight = 0;
	long N3_weight = 0;
	for (Vertex *a: N1) {
		N1_weight += a->permanent ? G.V.size() : a->weight;
	}
	if (N1_weight < k)
		goto cleanup;

	find_neighbourhood_helper(N1, N2);
	for (Vertex *a: N2) {
		N2_weight += a->permanent ? G.V.size() : a->weight;
	}
	if (N2_weight < k)
		goto cleanup;

	find_neighbourhood_helper(N2, N3);
	for (Vertex *a: N3) {
		N3_weight += a->permanent ? G.V.size() : a->weight;
	}
	if (N3_weight < k)
		goto cleanup;

	for (Vertex *a: N1) {
		a->marked = false;
		a->dist = 1;
	}
	for (Vertex *a: N2) {
		a->marked = false;
		a->dist = 2;
	}
	for (Vertex *a: N3) {
		a->marked = false;
		a->dist = 3;
	}


	#define F_IN(a) a->fvert[0]
	#define F_OUT(a) a->fvert[1]

	// setup type 1 flow

	for (Vertex *a: N1) {
		// create a "a_in" and "a_out" vertex connected by a single edge
		a->fvert.push_back(create_vertex(Gf));
		a->fvert.push_back(create_vertex(Gf));
		long weight = a->permanent ? G.V.size() : a->weight;
		//create_edge(Gf, F_IN(a), F_OUT(a), weight, nullptr);
		// connect to source
		create_edge(Gf, s, F_OUT(a), weight, nullptr);
	}
	for (Vertex *a: N2) {
		// create a "a_in" and "a_out" vertex connected by a single edge
		a->fvert.push_back(create_vertex(Gf));
		a->fvert.push_back(create_vertex(Gf));
		long weight = a->permanent ? G.V.size() : a->weight;
		create_edge(Gf, F_IN(a), F_OUT(a), weight, nullptr);
	}
	for (Vertex *a: N3) {
		a->fvert.push_back(create_vertex(Gf));
		// connect to sink
		long weight = a->permanent ? G.V.size() : a->weight;
		create_edge(Gf, F_IN(a), t, weight, nullptr);
	}

	for (Vertex *a: N2) {
		long weight_a = a->permanent ? G.V.size() : a->weight;
		for (auto edge: a->edges) {
			Vertex *b = edge.first;
			Edge *e = edge.second;
			if (b->dist != 1)
				continue;
			long weight_b = a->permanent ? G.V.size() : a->weight;

			create_edge(Gf, F_OUT(b), F_IN(a), min(weight_a, weight_b), e);
		}
	}
	for (Vertex *a: N3) {
		long weight_a = a->permanent ? G.V.size() : a->weight;
		for (auto edge: a->edges) {
			Vertex *b = edge.first;
			Edge *e = edge.second;
			if (b->dist != 2)
				continue;
			long weight_b = a->permanent ? G.V.size() : a->weight;

			create_edge(Gf, F_OUT(b), F_IN(a), min(weight_a, weight_b), e);
		}
	}

	flow = compute_max_flow(Gf, s, t);
	assert(flow <= min(N1_weight, min(N2_weight, N3_weight)));

	if (flow >= k + 1) {
		add_to_vds(G, k, v);
		success = true;
		goto cleanup;
	}



cleanup:
	for (Vertex *a: N3) {
		a->marked = false;
		a->dist = 0;
		a->fvert.clear();
	}
	for (Vertex *a: N2) {
		a->marked = false;
		a->dist = 0;
		a->fvert.clear();
	}
	for (Vertex *a: N1) {
		a->marked = false;
		a->dist = 0;
		a->fvert.clear();
	}
	v->marked = false;
	v->dist = 0;

	return success;
}

bool tc_kplus_disjoint(Graph &G, long long &k) {
	bool success = false;
	vector<Vertex *> V_tmp(G.V);

	for (Vertex *v: V_tmp) {
		if (k <= 0)
			break;
		if (v->deg == 0)
			continue;

		success = tc_kplus_disjoint_single(G, k, v) ? true : success;
	}

	return success;
}
